from unittest.mock import Mock

import pytest
from tests.test_connect import RAY_COMMIT, RAY_VERSION

from anyscale.utils.ray_version_checker import check_required_ray_version


@pytest.mark.parametrize(
    "version,commit",
    [
        pytest.param("1.1.0", "fake_commit", id="BothWrong"),
        pytest.param(RAY_VERSION, RAY_COMMIT[2:], id="BadCommit"),
        pytest.param("1.0.0", RAY_COMMIT, id="BadVersion"),
    ],
)
def test_invalid_check_required_ray_version(version: str, commit: str) -> None:
    with pytest.raises(ValueError):
        check_required_ray_version(
            Mock(),
            ray_version=RAY_VERSION,
            ray_commit=RAY_COMMIT,
            required_ray_commit=commit,
            required_ray_version=version,
            ignore_version_check=False,
        )
    check_required_ray_version(
        Mock(),
        ray_version=RAY_VERSION,
        ray_commit=RAY_COMMIT,
        required_ray_commit=commit,
        required_ray_version=version,
        ignore_version_check=True,
    )


@pytest.mark.parametrize("ignore", [False, True])
def test_valid_check_required_ray_version(ignore) -> None:
    check_required_ray_version(
        Mock(),
        RAY_VERSION,
        RAY_COMMIT,
        RAY_VERSION,
        RAY_COMMIT,
        ignore_version_check=ignore,
    )



from contextlib import contextmanager
import os
import time
from typing import Any, Dict, Optional, Tuple

import click

from anyscale.background.job_runner import _run_kill_child
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.app_config_config_schema import (
    AppConfigConfigSchema,
)
from anyscale.client.openapi_client.models.baseimagesenum import BASEIMAGESENUM
from anyscale.client.openapi_client.models.compute_node_type import ComputeNodeType
from anyscale.client.openapi_client.models.worker_node_type import WorkerNodeType
from anyscale.controllers.base_controller import BaseController
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)
from anyscale.utils.cloud_utils import get_default_cloud
from anyscale.utils.ray_version_checker import (
    detect_python_minor_version,
    detect_ray_version,
)


DEFAULT_RAY_VERSION = "1.6.0"


def run_command_on_all_nodes(command: str, timeout: Optional[float]):
    class Runner:
        """
        This class is similar to BackgroundJobRunner from job_runner.py.
        The version of run_background_job in BackgroundJobRunner takes an additional BackgroundJobContext argument
        that requires a creator_db_id and sets its own runtime environment.
        This simpler version does not.

        TODO(mattweber): If we ever release this, we will have to creator_db_id, otherwise this won't show in the UI
        """

        def run_background_job(self, command: str) -> None:
            # Update the context with the runtime env uris
            env_vars = {
                "PYTHONUNBUFFERED": "1",  # Make sure python subprocess streams logs https://docs.python.org/3/using/cmdline.html#cmdoption-u
            }
            env = {**os.environ, **env_vars}

            try:
                # TODO(mattweber): Once the publicly named run_kill_child is
                # available on product nodes, remove the underscore on this function.
                _run_kill_child(command, shell=True, check=True, env=env)  # noqa
            finally:
                # allow time for any logs to propogate before the task exits
                time.sleep(1)

    # import ray dependencies here to avoid making ray a general dependency for
    # the CLI
    import ray
    from ray.util.placement_group import placement_group

    nodes = ray.nodes()
    bundles = [{"CPU": 1} for node in nodes]
    pg = placement_group(bundles, strategy="STRICT_SPREAD")
    ray.get(pg.ready(), timeout=timeout)

    runners = [
        ray.remote(Runner)
        .options(  # type: ignore
            placement_group=pg, placement_group_bundle_index=i
        )
        .remote()
        for i, x in enumerate(nodes)
    ]
    futures = [runner.run_background_job.remote(command) for runner in runners]
    ray.get(futures)


class ConfigController(BaseController):
    """
    This controller powers functionalities related to Anyscale configurations such as
    cluster environments for compute configs.
    """

    def __init__(
        self, log: BlockLogger = BlockLogger(), initialize_auth_api_client: bool = True
    ):
        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.sdk_client = self.anyscale_api_client

        self.log = log

    def convert_cluster_yaml(
        self,
        cloud_name: Optional[str],
        cluster_yaml: Dict[str, Any],
        ml: bool,
        gpu: bool,
    ) -> Tuple[AppConfigConfigSchema, ClusterComputeConfig]:
        cloud_id, cloud_name = get_default_cloud(self.api_client, cloud_name)
        self.log.info(f"Using cloud: {cloud_name}")

        cluster_env = self._convert_cluster_env(cluster_yaml, ml, gpu)
        compute_config = self._convert_compute_config(cloud_id, cluster_yaml)
        return cluster_env, compute_config

    def create_cluster_env(self, name: str, cluster_env_config: Dict[str, Any]) -> None:
        config = AppConfigConfigSchema(**cluster_env_config)
        self.sdk_client.create_cluster_environment(
            {"name": name, "config_json": config.to_dict()}
        )

    def create_compute_config(
        self, name: str, compute_config: Dict[str, Any], anonymous: bool
    ) -> None:
        cluster_compute_config = ClusterComputeConfig(**compute_config)
        self.sdk_client.create_cluster_compute(
            {"name": name, "config": cluster_compute_config, "anonymous": anonymous}
        )

    def setup_dev_ray(
        self,
        working_dir: str,
        cluster_name: str,
        cluster_env: Optional[str],
        cluster_compute: Optional[str],
        timeout: Optional[float],
    ):
        # import ray dependencies here to avoid making ray a general dependency for
        # the CLI
        import ray

        optional_args = {}
        if cluster_env:
            optional_args["cluster_env"] = cluster_env
        if cluster_compute:
            optional_args["cluster_compute"] = cluster_compute

        ray.init(
            f"anyscale://{cluster_name}",
            runtime_env={"working_dir": working_dir},
            **optional_args,
        )
        run_command_on_all_nodes(
            "cp -RT ray /tmp/dev-ray; python /tmp/dev-ray/python/ray/setup-dev.py -y",
            timeout,
        )

    def _convert_node_config_aws(
        self, name: str, node_type_config: Dict[str, Any], is_head_node: bool = False,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Returns a tuple
        Index 0: a dictionary in the shape of either a ComputeNodeType or a WorkerNodeType depending on
        if `is_head_node` is True.
        Index 1: a dictionary of additional aws options which will be applied globally.
        """
        with self._warn_if_fail("Instance type is required", is_error=True):
            node_config = node_type_config["node_config"]
            instance_type = node_config["InstanceType"]

        min_workers = None
        max_workers = None
        use_spot = None

        if not is_head_node:
            min_workers = node_type_config.get("min_workers")
            max_workers = node_type_config.get("max_workers")
            market_type = node_config.get("InstanceMarketOptions", {}).get("MarketType")
            use_spot = market_type == "spot"

        # No need to handle resources. They will be filled in based on the instance type names

        output_node_config = {
            "name": name,
            "instance_type": instance_type,
            **({"min_workers": min_workers} if min_workers is not None else {}),
            **({"max_workers": max_workers} if max_workers is not None else {}),
            **({"use_spot": use_spot} if use_spot is not None else {}),
        }

        supported_option_keys = [
            "BlockDeviceMappings",
            "IamInstanceProfile",
            "TagSpecifications",
            "NetworkInterfaces",
        ]

        aws_options = {
            key: value
            for key, value in node_config.items()
            if key in supported_option_keys
        }

        return output_node_config, aws_options

    def _convert_compute_config(
        self, cloud_id: str, cluster_yaml: Dict[str, Any]
    ) -> ClusterComputeConfig:
        with self._warn_if_fail("Provider info is required", is_error=True):
            provider_config = cluster_yaml["provider"]
            provider_type = provider_config["type"]
            region = provider_config["region"]

        if provider_type != "aws":
            raise click.ClickException("Only AWS cluster yamls are supported for now.")

        # Provider configs
        availability_zone = provider_config.get("availability_zone")
        allowed_azs = availability_zone.split(",") if availability_zone else None

        # Available node types
        available_node_types = cluster_yaml.get("available_node_types")
        head_node_type = cluster_yaml.get("head_node_type")
        if not available_node_types or not head_node_type:
            # TODO (aguo): Do we need to handle legacy node types anymore?
            raise click.ClickException(
                "Available_node_types and head_node_type fields are required."
            )
        else:
            head_node = None
            worker_nodes = []
            global_aws_options: Dict[str, Any] = {}
            for name, node_type_config in available_node_types.items():
                if name == head_node_type:
                    node_config, aws_options = self._convert_node_config_aws(
                        name, node_type_config, is_head_node=True
                    )
                    head_node = ComputeNodeType(**node_config)
                else:
                    node_config, aws_options = self._convert_node_config_aws(
                        name, node_type_config
                    )
                    worker_nodes.append(WorkerNodeType(**node_config))

                global_aws_options = {**global_aws_options, **aws_options}

            if not head_node:
                raise click.BadArgumentUsage(
                    "available_node_types is missing the node_config for the head_node_type."
                )

            if len(global_aws_options):
                global_aws_options_keys = ", ".join(global_aws_options.keys())
                self.log.warning(
                    f"We squished the following configurations: ({global_aws_options_keys}) "
                    "from each node type config and set them globally across all node types."
                )

                if "IamInstanceProfile" in global_aws_options:
                    self.log.warning(
                        "WARNING: IamInstanceProfile found in cluster yaml. In most cases, "
                        "overriding this is not necessary because the anyscale cloud "
                        "that was selected will set up the correct iams for each node."
                    )

        return ClusterComputeConfig(
            cloud_id=cloud_id,
            max_workers=cluster_yaml.get("max_workers"),
            region=region,
            allowed_azs=allowed_azs,
            head_node_type=head_node,
            worker_node_types=worker_nodes,
            aws=global_aws_options,
        )

    def _convert_cluster_env(
        self, cluster_yaml: Dict[str, Any], ml: bool, gpu: bool,
    ) -> AppConfigConfigSchema:
        self.log.info(
            "Generating cluster environment based on local python and ray versions"
        )

        base_image = self._select_base_image(ml, gpu)

        post_build_commands = []
        setup_commands = cluster_yaml.get("setup_commands", [])
        if len(setup_commands):
            self.log.warning(
                "WARNING: Setup commands detected! These have been converted "
                "to post-build commands for your cluster env.\n\n"
                "IMPORTANT:\n"
                "1. If your setup commands is installing pip packages, "
                "please add those pip packages to the `python` list in your cluster env instead.\n"
                "2. If your setup commands is installing a ray version, make sure to install that "
                "same ray version locally."
            )
            post_build_commands = setup_commands
        head_setup_commands = cluster_yaml.get("head_setup_commands", [])
        worker_setup_commands = cluster_yaml.get("worker_setup_commands", [])
        if len(head_setup_commands) or len(worker_setup_commands):
            self.log.warning(
                "WARNING: Head or worker specific setup commands detected! These will not be "
                "converted in your cluster environment because the same cluster environment will "
                "be installed on the head and worker nodes. Please move these to the post build "
                "commands of the cluster environment to run on all nodes."
            )
        initialization_commands = cluster_yaml.get("initialization_commands", [])
        if len(initialization_commands):
            self.log.warning(
                "WARNING: Initialization commands detected! These will not be converted in your "
                "cluster environment because commands in Anyscale can only be run inside the cluster "
                "environment, whereas initialization commands are run in open source outside "
                "the docker container. Please move these to the post build commands of the "
                "cluster environment to run on all nodes."
            )

        if len(cluster_yaml.get("file_mounts", {})) or len(
            cluster_yaml.get("cluster_synced_files", {})
        ):
            self.log.warning(
                "WARNING: file_mounts or cluster_synced_files detected! These are not converted into a cluster env. "
                "If you need to upload local files into your cluster, please use a runtime environment."
            )

        return AppConfigConfigSchema(
            base_image=base_image,
            post_build_cmds=post_build_commands,
            python={"pip_packages": [], "conda_packages": []},
            debian_packages=[],
            env_vars={},
        )

    def _select_base_image(self, ml: bool, gpu: bool) -> str:
        python_version = detect_python_minor_version()
        ray_version, ray_commit = detect_ray_version()
        self.log.info(
            f"Detected python version: {python_version}. Detected ray version: {ray_version} and ray commit: {ray_commit}"
        )

        if "dev0" in ray_version:
            self.log.warning(
                "There is no built-in base images for nightly versions of Ray. Selecting "
                f"{DEFAULT_RAY_VERSION} as default. If you need to use a newer version of Ray, please install "
                "that Ray version as a post-build command."
            )
            ray_version = DEFAULT_RAY_VERSION

        return str(
            getattr(
                BASEIMAGESENUM,
                self._find_enum_name(python_version, ray_version, ml, gpu),
            )
        )

    def _find_enum_name(
        self, python_version: str, ray_version: str, ml: bool, gpu: bool
    ) -> str:
        enum_name = "ANYSCALE_RAY"
        if ml:
            enum_name += "_ML"

        enum_name += f"_{ray_version.replace('.', '_')}"

        major_version, minor_version = python_version.split(".")
        enum_name += f"_PY{major_version}{minor_version}"

        if gpu:
            enum_name += "_GPU"

        return enum_name

    @contextmanager
    def _warn_if_fail(self, message: str, is_error: bool = False) -> Any:
        try:
            yield
        except KeyError as e:
            if is_error:
                raise click.ClickException(message) from e
            else:
                self.log.warning(message)



import os
import re
import subprocess
import time
from typing import Any, Dict, Optional, Union
import uuid

import yaml

from anyscale.background.job_context import BackgroundJob, BackgroundJobContext
from anyscale.sdk.anyscale_client.sdk import AnyscaleSDK
from anyscale.shared_anyscale_utils.conf import (
    ANYSCALE_BACKGROUND_JOB_CONTEXT,
    BACKGROUND_ACTOR_NAME,
)


class BackgroundJobRunner:
    """
    This class is an actor that runs a shell command on the head node for an anyscale cluster.
    This class will:
    1. Pass a BackgroundJobContext as an environment variable
    2. execute the command in a subprocess (and stream logs appropriately)
    3. Gracefully exit when the command is complete
    """

    def run_background_job(
        self, command: str, self_handle: Any, context: BackgroundJobContext,
    ) -> None:
        import ray

        # Update the context with the runtime env uris
        uris = [u for u in ray.get_runtime_context().runtime_env["uris"]]
        context.runtime_env_uris = uris
        env_vars = {
            "PYTHONUNBUFFERED": "1",  # Make sure python subprocess streams logs https://docs.python.org/3/using/cmdline.html#cmdoption-u
            "RAY_ADDRESS": "anyscale://",  # Make sure that internal ray.init has an anyscale RAY_ADDRESS
            ANYSCALE_BACKGROUND_JOB_CONTEXT: context.to_json(),
        }
        env = {**os.environ, **env_vars}

        try:
            run_kill_child(command, shell=True, check=True, env=env)  # noqa
        finally:
            # allow time for any logs to propogate before the task exits
            time.sleep(1)

            self_handle.stop.remote()

    def stop(self) -> None:
        import ray

        ray.actor.exit_actor()


def _parse_runtime_env(
    env: Optional[Union[str, Dict[str, Any]]] = None
) -> Dict[str, Any]:
    """
    If `env` is a string, it represents a YAML file. Load the file and return the dict.
    Otherwise, return the dict that is passed in, or {} if nothing is passed
    """
    if isinstance(env, str):
        with open(env, "r") as stream:
            try:
                return yaml.safe_load(stream)  # type: ignore
            except yaml.YAMLError as e:
                raise ValueError(
                    f"Failed while attempting to load a runtime env yaml spec into a runtime environment dictionary. Please ensure that the file {env} is specified correctly",
                    e,
                )
    else:
        return env or {}


# TODO(mattweber): Once this code has been deployed in a CLI release
# that we can expect to be available on product clusters, migrate all
# use of _run_kill_child to this function.
def run_kill_child(
    *popenargs, input=None, timeout=None, check=False, **kwargs
) -> subprocess.CompletedProcess:
    return _run_kill_child(
        *popenargs, input=input, timeout=timeout, check=check, **kwargs
    )


# TODO(mattweber): This is a public function despite the underscore.
# Renaming this function in this PR is problematic for setup_dev_ray in
# config_controller because it needs to import this function
# for use in a remote function.
def _run_kill_child(
    *popenargs, input=None, timeout=None, check=False, **kwargs
) -> subprocess.CompletedProcess:
    """
    This function is a fork of subprocess.run with fewer args.
    The goal is to create a child subprocess that is GUARANTEED to exit when the parent exits
    This is accomplished by:
    1. Making sure the child is the head of a new process group
    2. Create a third "Killer" process that is responsible for killing the child when the parent dies
    3. Killer process checks every second if the parent is dead.
    4. Killing the entire process group when we want to kill the child

    Arguments are the same as subprocess.run
    """
    # Start new session ensures that this subprocess starts as a new process group
    with subprocess.Popen(start_new_session=True, *popenargs, **kwargs) as process:
        parent_pid = os.getpid()
        child_pid = process.pid
        child_pgid = os.getpgid(child_pid)

        # Open a new subprocess to kill the child process when the parent process dies
        # kill -s 0 parent_pid will succeed if the parent is alive.
        # If it fails, SIGKILL the child process group and exit
        subprocess.Popen(
            f"while kill -s 0 {parent_pid}; do sleep 1; done; kill -9 -{child_pgid}",
            shell=True,
            # Suppress output
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        try:
            stdout, stderr = process.communicate(input, timeout=timeout)
        except:  # noqa      (this is taken from subprocess.run directly)
            # Including KeyboardInterrupt, communicate handled that.
            process.kill()
            # We don't call process.wait() as .__exit__ does that for us.
            raise

        retcode = process.poll()
        if check and retcode:
            raise subprocess.CalledProcessError(
                retcode, process.args, output=stdout, stderr=stderr
            )
    return subprocess.CompletedProcess(process.args, retcode or 0, stdout, stderr)


def _generate_job_name(command: str) -> str:
    """
    Escape the command string to get the job name, and replace spaces and periods with _
    """
    return re.sub(r"\W+", "", re.sub(r"[\. ]", "_", command))


def _wait_for_db_id(
    sdk: AnyscaleSDK, cluster_id: str, ray_job_id: str, timeout_secs=20
) -> Any:
    for _ in range(timeout_secs):
        time.sleep(1)
        jobs = sdk.search_jobs(
            {"ray_job_id": ray_job_id, "cluster_id": cluster_id}
        ).results

        if len(jobs) > 0:
            # This should be the db job
            return jobs[0]

    return None


class JobNotFoundError(RuntimeError):
    pass


def run(
    command: str,
    runtime_env: Optional[Union[str, Dict[str, Any]]] = None,
    address: Optional[str] = None,
) -> BackgroundJob:
    """
    This function lets you execute a shell command in the background on a ray cluster, in a specific environment

    This function returns a BackgroundJobContext object that let's you wait for or kill the scheduled job

    :param command: This is the first param
    @param param2: this is the second param

    Args:
        command: The shell command to run. Eg. "python script.py"
        runtime_env: The OSS Ray runtime environment in which to run the command. Should be a dict or a path.
            Please see the Ray runtime env documentation for the format
        address: The address of the cluster to run the command on. Must start with 'anyscale://'

    Returns:
        BackgroundJob object. You can use this object to wait for the job, or to kill it

    Examples:
        >>> script.py
        >>> # connect as normal
        >>> ray.init("anyscale://")

        >>> runner.py
        >>> ctx = run("python script.py", runtime_env="env.yaml", address="anyscale://")
        >>> ctx.wait() # wait for the job
        >>> ctx.kill() # kill the job
    """
    # lazy imports
    # ray is required to run in the background
    import ray

    from anyscale import ClientBuilder

    # handle optional args
    address = address or os.getenv("RAY_ADDRESS") or "anyscale://"
    runtime_env = _parse_runtime_env(runtime_env)
    namespace = f"bg_{str(uuid.uuid4())}"

    if not address.startswith("anyscale://"):
        raise ValueError(
            "Anyscale run can only be used to connect to anyscale clusters"
        )

    _, inner_address = address.split("://", maxsplit=1)

    # This is just setting a default, and will be overwritten by the inner job name
    outer_job_name = _generate_job_name(command)

    # Connect the outer ray client.
    # Start an outer job that will be used to ship the users file system
    client = ClientBuilder(inner_address)
    logger = client._log
    logger.info(
        f"Running in namespace: {namespace}. You can use this namespace as a reference to access this background job in the future."
    )
    # Enable this job's metadata to be overwritten by any children jobs
    client._bg_set_outer_overwritable()
    client_context = (
        client.env(runtime_env).job_name(outer_job_name).namespace(namespace).connect()
    )

    # Construct the actor to run the inner job in
    BackgroundJobRunnerActor = ray.remote(BackgroundJobRunner)  # noqa
    actor = BackgroundJobRunnerActor.options(
        lifetime="detached", name=BACKGROUND_ACTOR_NAME
    ).remote()

    # grab the creator id from the job config of the outer job
    creator_db_id = client._job_config.metadata["creator_id"]
    outer_job_id = ray.get_runtime_context().job_id.hex()

    # generate a random cross reference id to identify the created job later
    context = BackgroundJobContext(
        creator_db_id=creator_db_id,
        original_command=command,
        namespace=namespace,
        # This will be populated in the actor
        runtime_env_uris=[],
        parent_ray_job_id=outer_job_id,
    )

    # Kick off the job
    remote_ref = actor.run_background_job.remote(
        command=command, self_handle=actor, context=context
    )

    # Wait for the Job to be populated in the db table and then fetch it and grab it's id
    db_job = _wait_for_db_id(
        client._anyscale_sdk,
        cluster_id=client_context.anyscale_cluster_info.cluster_id,
        ray_job_id=outer_job_id,
    )
    if not db_job:
        raise JobNotFoundError(
            "Failed to find the corresponding job in the Anyscale DB. This job may not have been scheduled."
        )

    return BackgroundJob(actor, remote_ref, context, client_context, db_job)



"""NodeProvider classes that add product-specific behaviour."""

import base64
import copy
import inspect
import json
import logging
import time
from types import ModuleType
from typing import Any, cast, Dict, List, Optional

from openapi_client.rest import ApiException
from ray import cloudpickle
from ray.autoscaler._private.aws.node_provider import to_aws_format
from ray.autoscaler._private.command_runner import DockerCommandRunner, SSHCommandRunner
from ray.autoscaler._private.util import hash_launch_conf
from ray.autoscaler.command_runner import CommandRunnerInterface
from ray.autoscaler.tags import (
    STATUS_UP_TO_DATE,
    TAG_RAY_CLUSTER_NAME,
    TAG_RAY_NODE_STATUS,
)
from urllib3.exceptions import HTTPError  # type: ignore
import wrapt

from anyscale.authenticate import get_auth_api_client


# This try block exists because
# 1. This module is passed to the autoscaler on the head node as an external node provider
# 2. Import paths are different for different versions of ray
try:
    from ray.autoscaler._private.providers import _get_node_provider, _NODE_PROVIDERS
    from ray.autoscaler.node_provider import NodeProvider
except ModuleNotFoundError:
    from ray.autoscaler.aws.node_provider import NodeProvider
    from ray.autoscaler.node_provider import (
        get_node_provider as _get_node_provider,
        NODE_PROVIDERS as _NODE_PROVIDERS,
    )

from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.add_instance_pool_member import (
    AddInstancePoolMember,
)
from anyscale.client.openapi_client.models.create_nodes_options import (
    CreateNodesOptions,
)
from anyscale.client.openapi_client.models.nodes_options import NodesOptions
from anyscale.client.openapi_client.models.non_terminated_nodes_options import (
    NonTerminatedNodesOptions,
)
from anyscale.client.openapi_client.models.query_pool_size import QueryPoolSize
from anyscale.client.openapi_client.models.request_instance_pool_member import (
    RequestInstancePoolMember,
)
from anyscale.client.openapi_client.models.set_node_tags_options import (
    SetNodeTagsOptions,
)
from anyscale.client.openapi_client.models.wait_until_stopped_options import (
    WaitUntilStoppedOptions,
)
from anyscale.shared_anyscale_utils.util import get_autoscaler_container_path


logger = logging.getLogger(__name__)
RETRY_COUNT = 5


def wrap_api_client_with_retries(api_client: DefaultApi) -> DefaultApi:
    @wrapt.decorator  # type: ignore
    def retrier(func: Any, instance: Any, args: List[Any], kwargs: Any) -> Any:
        for i in range(RETRY_COUNT):
            try:
                return func(*args, **kwargs)
            except (ConnectionResetError, HTTPError) as e:
                if i == (RETRY_COUNT - 1):
                    logger.error(f"{func.__name__} FAILED {RETRY_COUNT} times!")
                    raise e
                time.sleep(0.1 * i)
                continue

    for name, fn in inspect.getmembers(api_client):
        if inspect.ismethod(fn) and "with_http_info" not in name:
            setattr(api_client, name, retrier(fn))
    return api_client


class AnyscaleInstanceManagerNodeProvider:
    """NodeProvider for managing nodes with instance manager.
    """

    def __init__(self, provider_config: Dict[str, Any], cluster_name: str) -> None:
        self.provider_config = provider_config
        self.inner_provider_config = provider_config["inner_provider"]
        self.inner_provider = _get_node_provider(
            self.inner_provider_config, cluster_name
        )
        self.cloud_id = self.inner_provider_config["cloud_id"]
        cli_token = self.inner_provider_config["cli_token"]
        host = self.inner_provider_config["host"]
        self.api_client = wrap_api_client_with_retries(
            get_auth_api_client(cli_token=cli_token, host=host).api_client
        )
        self.is_provider_supported = self._is_provider_supported()

    def _is_provider_supported(self) -> bool:
        """Currently instance manager only suppport AWS."""
        # TODO(yifei) move checks from _configure_for_cloud to here
        return True

    def __getattr__(self, name: str) -> Any:
        """Implements inheritance from self.inner_provider."""
        return getattr(self.inner_provider, name)

    # Use the same methodology as AnyscalePoolingNodeProvider for the following
    # static methods
    @staticmethod
    def _get_inner_provider_class_from_config(cluster_config: Dict[str, Any]) -> Any:

        inner_provider_config = cluster_config["provider"]["inner_provider"]
        importer = _NODE_PROVIDERS.get(inner_provider_config["type"])
        inner_provider_cls = importer(inner_provider_config)
        return inner_provider_cls

    @staticmethod
    def bootstrap_config(cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        cluster_config = copy.deepcopy(cluster_config)
        inner_provider_cls = AnyscalePoolingNodeProvider._get_inner_provider_class_from_config(
            cluster_config
        )
        primary_provider = cluster_config["provider"]
        cluster_config["provider"] = cluster_config["provider"]["inner_provider"]

        cluster_config = inner_provider_cls.bootstrap_config(cluster_config)

        primary_provider["inner_provider"] = cluster_config["provider"]
        cluster_config["provider"] = primary_provider
        return cluster_config

    @staticmethod
    def fillout_available_node_types_resources(
        cluster_config: Dict[str, Any]
    ) -> Dict[str, Any]:
        cluster_config = copy.deepcopy(cluster_config)
        inner_provider_cls = AnyscalePoolingNodeProvider._get_inner_provider_class_from_config(
            cluster_config
        )

        primary_provider = cluster_config["provider"]
        cluster_config["provider"] = cluster_config["provider"]["inner_provider"]

        cluster_config = inner_provider_cls.fillout_available_node_types_resources(
            cluster_config
        )

        primary_provider["inner_provider"] = cluster_config["provider"]
        cluster_config["provider"] = primary_provider
        return cluster_config

    def non_terminated_nodes(self, tag_filters: Dict[str, str]) -> Any:
        if not self.is_provider_supported:
            return self.inner_provider.non_terminated_nodes(tag_filters)

        non_terminated_nodes_request = NonTerminatedNodesOptions(
            provider_config=self.inner_provider.provider_config,
            cluster_name=self.inner_provider.cluster_name,
            tag_filters=tag_filters,
        )
        resp = self.api_client.non_terminated_nodes_api_v2_instances_non_terminated_nodes_post(
            non_terminated_nodes_request
        )
        return [instance.instance_id for instance in resp.results]

    def is_running(self, node_id: str) -> Any:
        if not self.is_provider_supported:
            return self.inner_provider.is_running(node_id)

        resp = self.api_client.is_running_api_v2_instances_instance_id_is_running_get(
            node_id
        )
        return resp.result.is_running

    def is_terminated(self, node_id: str) -> Any:
        if not self.is_provider_supported:
            return self.inner_provider.is_terminated(node_id)

        resp = self.api_client.is_terminated_api_v2_instances_instance_id_is_terminated_get(
            node_id
        )
        return resp.result.is_terminated

    def node_tags(self, node_id: str) -> Any:
        if not self.is_provider_supported:
            return self.inner_provider.node_tags(node_id)

        resp = self.api_client.get_instance_api_v2_instances_instance_id_get(node_id)
        return resp.result.tags

    def external_ip(self, node_id: str) -> Any:
        if not self.is_provider_supported:
            return self.inner_provider.external_ip(node_id)

        resp = self.api_client.external_ip_api_v2_instances_instance_id_external_ip_get(
            node_id
        )
        return resp.result.external_ip

    def internal_ip(self, node_id: str) -> Any:
        if not self.is_provider_supported:
            return self.inner_provider.internal_ip(node_id)

        resp = self.api_client.internal_ip_api_v2_instances_instance_id_internal_ip_get(
            node_id
        )
        return resp.result.internal_ip

    def create_node_with_resources(
        self,
        node_config: Dict[str, Any],
        tags: Dict[str, str],
        count: int,
        resources: Dict[str, float],
    ) -> None:
        """Mirrors Ray's NodeProvider interface. Needed for compatibility with
        Ray Autoscaler version >=1.8.0

        Refer to
        https://github.com/anyscale/product/pull/8469
        https://github.com/ray-project/ray/blob/releases/1.8.0/python/ray/autoscaler/node_provider.py#L127
        https://github.com/ray-project/ray/blob/releases/1.8.0/python/ray/autoscaler/_private/node_launcher.py#L69
        """
        self.create_node(node_config, tags, count)

    def create_node(
        self, node_config: Dict[str, Any], tags: Dict[str, str], count: int
    ) -> None:
        if not self.is_provider_supported:
            self.inner_provider.create_node(node_config, tags, count)
            return

        create_node_request = CreateNodesOptions(
            provider_config=self.inner_provider.provider_config,
            cluster_name=self.inner_provider.cluster_name,
            node_config=node_config,
            tags=tags,
            count=count,
        )
        # Catch, process, and repropagate remote exceptions.
        remote_exc = None
        try:
            self.api_client.create_nodes_api_v2_instances_post(create_node_request)
        except ApiException as api_exception:
            try:
                error_detail = json.loads(api_exception.body)
                ser_exc = error_detail["error"]["detail"]
                remote_exc = cloudpickle.loads(base64.b64decode(ser_exc))
            except Exception as deserialization_error:
                logger.warn(
                    "create_node() received an ApiException, "
                    "but we were unable to deserialize it "
                    f"({deserialization_error}). "
                    "Raising the ApiException unchanged."
                )
                raise api_exception

        # Raise the unwrapped error away from the try/except clause
        # so the ApiException codegenned frames do not tag along.
        if remote_exc:
            logger.warn(
                "create_node() received an ApiException. "
                "Unwrapping and raising the original remote exception instead."
            )
            raise remote_exc

    def set_node_tags(self, node_id: str, tags: Dict[str, str]) -> None:
        if not self.is_provider_supported:
            self.inner_provider.set_node_tags(node_id, tags)
            return

        set_node_tags_request = SetNodeTagsOptions(
            provider_config=self.inner_provider.provider_config,
            cluster_name=self.inner_provider.cluster_name,
            instance_id=node_id,
            tags=tags,
        )

        self.api_client.set_node_tags_api_v2_instances_set_tags_post(
            set_node_tags_request
        )

    def terminate_node(self, node_id: str) -> None:
        if not self.is_provider_supported:
            self.inner_provider.terminate_node(node_id)
            return

        terminate_nodes_request = NodesOptions(
            provider_config=self.inner_provider.provider_config,
            cluster_name=self.inner_provider.cluster_name,
            instance_ids=[node_id],
        )
        self.api_client.terminate_nodes_api_v2_instances_terminate_nodes_post(
            terminate_nodes_request
        )

    def terminate_nodes(self, node_ids: List[str]) -> None:
        if not self.is_provider_supported:
            self.inner_provider.terminate_nodes(node_ids)
            return

        terminate_nodes_request = NodesOptions(
            provider_config=self.inner_provider.provider_config,
            cluster_name=self.inner_provider.cluster_name,
            instance_ids=node_ids,
        )
        self.api_client.terminate_nodes_api_v2_instances_terminate_nodes_post(
            terminate_nodes_request
        )

    def wait_until_stopped(self, node_id: str) -> None:
        wait_until_stopped_request = WaitUntilStoppedOptions(
            cloud_id=self.cloud_id, instance_id=node_id,
        )
        self.api_client.wait_until_stopped_api_v2_instances_wait_until_stopped_post(
            wait_until_stopped_request
        )

    def start_nodes(self, node_ids: List[str]) -> None:
        if not self.is_provider_supported:
            self.inner_provider.start_nodes(node_ids)
            return

        start_nodes_request = NodesOptions(
            provider_config=self.inner_provider.provider_config,
            cluster_name=self.inner_provider.cluster_name,
            instance_ids=node_ids,
        )
        self.api_client.start_nodes_api_v2_instances_start_nodes_post(
            start_nodes_request
        )

    def stop_nodes(self, node_ids: List[str]) -> None:
        if not self.is_provider_supported:
            self.inner_provider.stop_nodes(node_ids)
            return

        stop_nodes_request = NodesOptions(
            provider_config=self.inner_provider.provider_config,
            cluster_name=self.inner_provider.cluster_name,
            instance_ids=node_ids,
        )
        self.api_client.stop_nodes_api_v2_instances_stop_nodes_post(stop_nodes_request)

    def get_node_id(
        self, ip_address: str, use_internal_ip: bool = False
    ) -> Optional[str]:
        if not self.is_provider_supported:
            return cast(
                str, self.inner_provider.get_node_id(ip_address, use_internal_ip)
            )

        if use_internal_ip:
            resp = self.api_client.get_node_id_by_internal_ip_api_v2_instances_internal_ip_get_node_id_by_internal_ip_get(
                ip_address
            )
        else:
            resp = self.api_client.get_node_id_by_external_ip_api_v2_instances_external_ip_get_node_id_by_external_ip_get(
                ip_address
            )

        if not resp.result.instance_id:
            raise ValueError("IP {ip_address} not found.")
        return cast(str, resp.result.instance_id)

    def get_command_runner(
        self,
        log_prefix: str,
        node_id: str,
        auth_config: Dict[str, Any],
        cluster_name: str,
        process_runner: ModuleType,
        use_internal_ip: bool,
        docker_config: Optional[Dict[str, Any]] = None,
    ) -> CommandRunnerInterface:
        """Copied from aws node provider. Need to pass
        """
        common_args = {
            "log_prefix": log_prefix,
            "node_id": node_id,
            "provider": self,
            "auth_config": auth_config,
            "cluster_name": cluster_name,
            "process_runner": process_runner,
            "use_internal_ip": use_internal_ip,
        }
        if docker_config and docker_config["container_name"] != "":
            return DockerCommandRunner(docker_config, **common_args)
        else:
            return SSHCommandRunner(**common_args)

    def prepare_for_head_node(self, cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        cluster_config = copy.deepcopy(cluster_config)

        def delete_credentials_from_config(provider_config: Dict[str, Any]) -> None:
            """
            Context (as of 2020-12-11):
            * In order to allow an autoscaler running in our account to access a
              customer's account we have to place credentials into the cluster
              config so that the NodeProvider can find them.
            * The credentials we generate to access customer accounts are
              time-limited to expire after an hour.
            * The autoscaler will copy the provided cluster_config over to the
              head node for it to use for launching worker nodes
            * If there are credentials explicitly provided to boto3 (via the
              NodeProvider), it will not try to find alternatives if those are
              invalid.
            * The head node is launched with a role that allows it to launch
              more nodes, it does not need credentials explicitly passed to it.
            This creates a problem where we put time-limited credentials in the
            autoscaler config, which is then copied to the head node. The head
            node will then use those credentials for as long as they're valid to
            manage worker nodes. However once they expire the head node will no
            longer be able to make API calls using the explicitly provided
            credentials. 😞
            To work around this (until we have a better way to configure the
            autoscaler's API clients) is to delete any credentials we can find
            in the cluster_config before it gets handed to the node.
            """
            provider_config.pop("aws_credentials", None)

            # NodePool provider
            if "inner_provider" in provider_config:
                delete_credentials_from_config(provider_config["inner_provider"])

        delete_credentials_from_config(cluster_config.get("provider", {}))

        file_mounts = cluster_config.get("file_mounts", {})
        for remote, local in file_mounts.items():
            file_mounts[remote] = get_autoscaler_container_path(local)

        return cluster_config


class AnyscalePoolingNodeProvider(AnyscaleInstanceManagerNodeProvider):
    """NodeProvider wrapper for injecting product-specific functionality.

    https://docs.google.com/document/d/15goOexCiGkbzz7tUbILMc10Gmjf86PiyCtIl75-xGUw/edit
    """

    POOLED_INSTANCE_TAG = "anyscale-launched-from-pool"
    INSTANCE_REUSE_TAG = "anyscale-instance-reuse-id"

    def _get_from_configs(self, key: str, default: Any = KeyError) -> Any:
        """Searches both the main and inner provider configs for a key."""

        if key in self.provider_config:
            return self.provider_config[key]
        elif key in self.inner_provider_config:
            return self.inner_provider_config[key]
        elif default == KeyError:
            raise KeyError
        else:
            return default

    def create_node(
        self, node_config: Dict[str, Any], tags: Dict[str, str], count: int
    ) -> None:
        # Calculate the instance reuse ID straight away
        # so that it can be stored in the node's tags
        # for later potential insertion into the pool.
        #
        # The reuse ID is currently the hash of the node config, minus user tags.
        config_to_hash = copy.deepcopy(node_config)
        if "TagSpecifications" in config_to_hash:
            del config_to_hash["TagSpecifications"]
        instance_reuse_id = hash_launch_conf(config_to_hash, dict())
        tags[self.INSTANCE_REUSE_TAG] = instance_reuse_id

        # Only handle AWS for now.
        # TODO(xcl): add support for non-aws instances.
        bypass_rapid_start = self._get_from_configs("bypass_rapid_start", None)
        pool_disabled = self._pool_limit() == 0
        # Pooling version (with checks).
        if not (bypass_rapid_start or pool_disabled):
            if self._is_aws_provider():
                for _ in range(count):
                    self._create_single_node_aws(node_config, tags)
                return

        # Non-pooling version.
        logger.info("Configuration bypasses Rapid Start; launching node normally.")
        super().create_node(node_config, tags, count)

    def _create_single_node_aws(
        self, node_config: Dict[str, Any], tags: Dict[str, str],
    ) -> None:
        """create_node implementation for when inner provider is AWS."""

        instance_pool_request = RequestInstancePoolMember(
            cloud_id=self.cloud_id, instance_reuse_id=tags[self.INSTANCE_REUSE_TAG],
        )

        try:
            instance_pool_member = self.api_client.request_instance_from_pool_api_v2_instancepools_request_instance_post(
                instance_pool_request
            ).result
            node_id = instance_pool_member.instance_id
        except ApiException:
            # We expect to get this exception if there is no instance
            # that matches this request.
            node_id = None

        # If we get a node_id back from the instance pool,
        # start this node.
        if node_id is not None:
            logger.info("Launching node from instance pool! ID {}".format(node_id))
            super().start_nodes([node_id])

            tags[self.POOLED_INSTANCE_TAG] = "true"

            # Various mandatory tag modification logic
            # that happens in OSS under create_node.
            tags = to_aws_format(tags)
            tags[TAG_RAY_CLUSTER_NAME] = self.cluster_name
            # Add the tags from the node config's TagSpecifications.
            if "TagSpecifications" in node_config:
                for tagspec in node_config["TagSpecifications"]:
                    if tagspec["ResourceType"] == "instance":
                        for tagpair in tagspec["Tags"]:
                            tags[tagpair["Key"]] = tagpair["Value"]

            logger.info("Setting node tags on reused instance: {}".format(str(tags)))
            super().set_node_tags(node_id, tags)

        # Otherwise, start a node normally.
        else:
            logger.info("Instance pool empty; launching node from EC2.")
            super().create_node(node_config, tags, count=1)

    def terminate_node(self, node_id: str) -> None:
        # Only handle AWS for now.
        # TODO(xcl): add support for non-aws instances.
        tags = self.node_tags(node_id)
        launch_hash = tags.get(self.INSTANCE_REUSE_TAG)

        # Check prerequisites and fallback to normal termination
        # if prerequisites are not met.
        if launch_hash is None:
            logger.warn(
                "No instance reuse tag found, so node is ineligible for pool. "
                "This should only happen if the node was launched using an older CLI version. "
            )
            super().terminate_node(node_id)
            return

        if not self._is_aws_provider():
            logger.warn(
                "Instance pool not yet supported for this cloud provider; node will not be added to pool."
            )
            super().terminate_node(node_id)
            return

        if self._room_in_pool() <= 0:
            logger.warn("No room in pool; node will not be added to pool.")
            super().terminate_node(node_id)
            return

        # Prerequisites are met here;
        # proceed with putting node in instance pool.
        self._prepare_node_for_idempotent_reuse(node_id)
        # TODO(xcl): Determine which tags should be cleared vs left uncleared for correctness here.

        super().stop_nodes([node_id])
        super().wait_until_stopped(node_id)

        logger.info("Sending node {} to the instance pool.".format(node_id))

        instance_reuse_id = launch_hash
        instance_pool_submission = AddInstancePoolMember(
            instance_id=node_id,
            cloud_id=self.cloud_id,
            instance_reuse_id=instance_reuse_id,
        )
        self.api_client.add_instance_to_pool_api_v2_instancepools_add_instance_post(
            instance_pool_submission
        )

    def _prepare_node_for_idempotent_reuse(self, node_id: str) -> None:
        """Does necessary resets on node to prepare it for reuse."""

        # Currently, we will explicitly not do anything.
        # The only potential issue right now is file mounts from previous sessions,
        # which should ideally be garbage collected (TODO (xcl)),
        # but this is not a high priority because we may also just
        # expire sessions in the pool after they've been reused a lot.

    def terminate_nodes(self, node_ids: List[str]) -> None:
        # TODO(xcl): add support for non-aws instances.
        if self._is_aws_provider():
            for node_id in node_ids:
                self.terminate_node(node_id)
        else:
            super().terminate_nodes(node_ids)

    def _is_aws_provider(self) -> bool:
        """Helper method; whether the inner provider is AWS or not."""

        # 1. The inner provider can be AWS directly.
        if self.inner_provider_config["type"] == "aws":
            return True

        # 2. If the inner provider is of type "external",
        # it can still represent an AWS provider,
        # only if the corresponding module is our AnyscaleAWSNodeProvider.
        if (
            self.inner_provider_config["type"] == "external"
            and self.inner_provider_config["module"]
            == "anyscale.autoscaler.aws.node_provider.AnyscaleAWSNodeProvider"
        ):
            return True

        return False

    def _pool_size(self) -> int:
        """Get this cloud's current pool size."""
        query = QueryPoolSize(cloud_id=self.cloud_id)
        pool_size = self.api_client.pool_size_api_v2_instancepools_pool_size_post(query)
        return int(pool_size)

    def _pool_limit(self) -> int:
        """Get this cloud's pool size limit."""

        pool_limit = self.api_client.get_cloud_api_v2_clouds_cloud_id_get(
            cloud_id=self.cloud_id
        ).result.config.max_stopped_instances

        return int(pool_limit)  # already int; just cast for typing

    def _room_in_pool(self) -> int:
        return self._pool_limit() - self._pool_size()


class AnyscaleExecNodeProvider(NodeProvider):  # type: ignore
    """A temporary class for making `anyscale exec` faster.

    Only used in the frontend CLI.
    """

    def __init__(self, provider_config: Dict[str, Any], cluster_name: str) -> None:
        super().__init__(provider_config, cluster_name)
        self.dns_address = provider_config["dns_address"]

    def non_terminated_nodes(self, tag_filters: List[str]) -> List[str]:
        return [self.dns_address]

    def external_ip(self, node_id: str) -> str:
        return node_id

    def node_tags(self, node_id: str) -> Dict[str, Any]:
        # Necessary because Ray requires that head node to have this tag set.
        return {TAG_RAY_NODE_STATUS: STATUS_UP_TO_DATE}



import subprocess
from typing import Any, Dict, List
from unittest.mock import call, Mock, patch

import pytest

from anyscale.authenticate import get_auth_api_client
from anyscale.autosync_heartbeat import (
    managed_autosync_session,
    perform_autopush_synchronization,
)


def test_managed_autosync_session() -> None:
    with patch.multiple(
        "anyscale.authenticate.AuthenticationBlock",
        _validate_api_client_auth=Mock(),
        _validate_credentials_format=Mock(),
    ):
        mock_api_client = get_auth_api_client(
            cli_token="fake_cli_token", host="fake_host"
        ).api_client
    mock_api_client.register_autosync_session_api_v2_autosync_sessions_post = Mock(
        return_value=Mock(result=Mock(id=1))
    )
    mock_api_client.deregister_autosync_session_api_v2_autosync_sessions_autosync_session_id_delete = Mock(
        return_value=None
    )

    with managed_autosync_session("id", mock_api_client):
        pass

    mock_api_client.register_autosync_session_api_v2_autosync_sessions_post.assert_called_once_with(
        "id"
    )
    mock_api_client.deregister_autosync_session_api_v2_autosync_sessions_autosync_session_id_delete.assert_called_once_with(
        1
    )


# Tests autosync without delete flag
def test_perform_autopush_synchronization() -> None:
    mock_proc = Mock()
    mock_proc.stdout = Mock()

    def readline(*args: List[Any], **kwargs: Dict[str, Any]) -> bytes:
        return b"30"

    mock_proc.stdout.readline.side_effect = readline

    class MockTimer:
        def __init__(self, time: int, func: Any):
            self._func = func

        def start(self) -> None:
            self._func()

    with patch("anyscale.autosync_heartbeat.subprocess.Popen") as mock_popen, patch(
        "anyscale.autosync_heartbeat.subprocess.check_output"
    ) as mock_check_output, patch(
        "anyscale.autosync_heartbeat.subprocess.run"
    ) as mock_run, patch(
        "anyscale.autosync_heartbeat.get_rsync_command"
    ) as mock_get_rsync_command, patch(
        "anyscale.autosync_heartbeat.Timer", new=MockTimer
    ), patch(
        "anyscale.autosync_heartbeat.sys.platform", new="darwin"
    ), patch(
        "anyscale.autosync_heartbeat.os.path.dirname"
    ) as mock_dir_name:
        mock_dir_name.return_value = "test"
        mock_popen.return_value.__enter__.return_value = mock_proc
        # Mock no deleted files
        mock_check_output.return_value = b""
        # Raise exception to exit out of infinite loop
        mock_run.side_effect = Exception()
        mock_get_rsync_command.return_value = (["rsync", "test"], {"test-env": "test"})

        with pytest.raises(Exception):
            perform_autopush_synchronization(
                ["ssh"],
                "project_folder",
                "test-user",
                "1.1.1.1",
                "target_folder",
                False,
            )

        mock_popen.assert_called_once_with(
            ["test/fswatch-darwin", "project_folder", "-o"],
            stdout=subprocess.PIPE,
            env={"DYLD_LIBRARY_PATH": "test"},
        )

        mock_get_rsync_command.assert_has_calls(
            [
                call(
                    ["ssh"],
                    "project_folder",
                    "test-user",
                    "1.1.1.1",
                    "target_folder",
                    delete=True,
                    dry_run=True,
                    rsync_exclude=[],
                    rsync_filter=[],
                ),
                call(
                    ["ssh"],
                    "project_folder",
                    "test-user",
                    "1.1.1.1",
                    "target_folder",
                    False,
                    rsync_exclude=[],
                    rsync_filter=[],
                ),
            ]
        )

        mock_run.assert_called_once_with(["rsync", "test"], env={"test-env": "test"})


# Tests autosync with delete flag
def test_perform_autopush_synchronization_with_delete_flag_on() -> None:
    mock_proc = Mock()
    mock_proc.stdout = Mock()

    def readline(*args: List[Any], **kwargs: Dict[str, Any]) -> bytes:
        return b"30"

    mock_proc.stdout.readline.side_effect = readline

    class MockTimer:
        def __init__(self, time: int, func: Any):
            self._func = func

        def start(self) -> None:
            self._func()

    with patch("anyscale.autosync_heartbeat.subprocess.Popen") as mock_popen, patch(
        "anyscale.autosync_heartbeat.subprocess.check_output"
    ) as mock_check_output, patch(
        "anyscale.autosync_heartbeat.subprocess.run"
    ) as mock_run, patch(
        "anyscale.autosync_heartbeat.get_rsync_command"
    ) as mock_get_rsync_command, patch(
        "anyscale.autosync_heartbeat.Timer", new=MockTimer
    ), patch(
        "anyscale.autosync_heartbeat.sys.platform", new="darwin"
    ), patch(
        "anyscale.autosync_heartbeat.os.path.dirname"
    ) as mock_dir_name:
        mock_dir_name.return_value = "test"
        mock_popen.return_value.__enter__.return_value = mock_proc
        # Raise exception to exit out of infinite loop
        mock_run.side_effect = Exception()
        mock_get_rsync_command.return_value = (["rsync", "test"], {"test-env": "test"})

        with pytest.raises(Exception):
            perform_autopush_synchronization(
                ["ssh"],
                "project_folder",
                "test-user",
                "1.1.1.1",
                "target_folder",
                True,
            )

        mock_popen.assert_called_once_with(
            ["test/fswatch-darwin", "project_folder", "-o"],
            stdout=subprocess.PIPE,
            env={"DYLD_LIBRARY_PATH": "test"},
        )

        mock_get_rsync_command.assert_called_once_with(
            ["ssh"],
            "project_folder",
            "test-user",
            "1.1.1.1",
            "target_folder",
            True,
            rsync_exclude=[],
            rsync_filter=[],
        )

        mock_check_output.assert_not_called()

        mock_run.assert_called_once_with(["rsync", "test"], env={"test-env": "test"})


# Tests autosync without delete flag
def test_perform_autopush_synchronization_with_exclude_and_filter() -> None:
    mock_proc = Mock()
    mock_proc.stdout = Mock()

    def readline(*args: List[Any], **kwargs: Dict[str, Any]) -> bytes:
        return b"30"

    mock_proc.stdout.readline.side_effect = readline

    class MockTimer:
        def __init__(self, time: int, func: Any):
            self._func = func

        def start(self) -> None:
            self._func()

    with patch("anyscale.autosync_heartbeat.subprocess.Popen") as mock_popen, patch(
        "anyscale.autosync_heartbeat.subprocess.check_output"
    ) as mock_check_output, patch(
        "anyscale.autosync_heartbeat.subprocess.run"
    ) as mock_run, patch(
        "anyscale.autosync_heartbeat.get_rsync_command"
    ) as mock_get_rsync_command, patch(
        "anyscale.autosync_heartbeat.Timer", new=MockTimer
    ), patch(
        "anyscale.autosync_heartbeat.sys.platform", new="darwin"
    ), patch(
        "anyscale.autosync_heartbeat.os.path.dirname"
    ) as mock_dir_name:
        mock_dir_name.return_value = "test"
        mock_popen.return_value.__enter__.return_value = mock_proc
        # Mock no deleted files
        mock_check_output.return_value = b""
        # Raise exception to exit out of infinite loop
        mock_run.side_effect = Exception()
        mock_get_rsync_command.return_value = (["rsync", "test"], {"test-env": "test"})

        with pytest.raises(Exception):
            perform_autopush_synchronization(
                ["ssh"],
                "project_folder",
                "test-user",
                "1.1.1.1",
                "target_folder",
                False,
                rsync_exclude=["**/.git", "**/.git/**"],
                rsync_filter=[".gitignore"],
            )

        mock_popen.assert_called_once_with(
            ["test/fswatch-darwin", "project_folder", "-o"],
            stdout=subprocess.PIPE,
            env={"DYLD_LIBRARY_PATH": "test"},
        )

        mock_get_rsync_command.assert_has_calls(
            [
                call(
                    ["ssh"],
                    "project_folder",
                    "test-user",
                    "1.1.1.1",
                    "target_folder",
                    delete=True,
                    dry_run=True,
                    rsync_exclude=["**/.git", "**/.git/**"],
                    rsync_filter=[".gitignore"],
                ),
                call(
                    ["ssh"],
                    "project_folder",
                    "test-user",
                    "1.1.1.1",
                    "target_folder",
                    False,
                    rsync_exclude=["**/.git", "**/.git/**"],
                    rsync_filter=[".gitignore"],
                ),
            ]
        )

        mock_run.assert_called_once_with(["rsync", "test"], env={"test-env": "test"})



import json
import logging
import os
import sys
from typing import Any

import click

from anyscale.cli_logger import BlockLogger
from anyscale.commands.anyscale_api.api_commands import anyscale_api
from anyscale.commands.cloud_commands import cloud_cli
from anyscale.commands.cluster_commands import cluster_cli
from anyscale.commands.cluster_compute_commands import cluster_compute_cli
from anyscale.commands.cluster_env_commands import cluster_env_cli
from anyscale.commands.config_commands import config_cli
from anyscale.commands.exec_commands import anyscale_exec
from anyscale.commands.job_commands import job_cli
from anyscale.commands.list_commands import list_cli
from anyscale.commands.migrate_commands import migrate_cli
from anyscale.commands.project_commands import (
    anyscale_clone,
    anyscale_init,
    project_cli,
)
from anyscale.commands.run_commands import anyscale_kill_job, anyscale_run
from anyscale.commands.service_commands import service_cli
from anyscale.commands.session_commands import (
    anyscale_autopush,
    anyscale_autosync,
    anyscale_fork,
    anyscale_pull,
    anyscale_push,
    anyscale_ssh,
    anyscale_start,
    anyscale_stop,
    anyscale_up,
)
from anyscale.commands.session_commands_hidden import session_cli
import anyscale.conf
from anyscale.util import init_sentry


logger = logging.getLogger(__file__)
logging.getLogger("botocore").setLevel(logging.CRITICAL)

log = BlockLogger()  # CLI Logger

if anyscale.conf.AWS_PROFILE is not None:
    logger.info("Using AWS profile %s", anyscale.conf.AWS_PROFILE)
    os.environ["AWS_PROFILE"] = anyscale.conf.AWS_PROFILE


class AliasedGroup(click.Group):
    # This is from https://stackoverflow.com/questions/46641928/python-click-multiple-command-names
    def get_command(self, ctx: Any, cmd_name: str) -> Any:
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(
    invoke_without_command=True,
    no_args_is_help=True,
    cls=AliasedGroup,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option(
    "--version",
    "-v",
    "version_flag",
    is_flag=True,
    default=False,
    help="Current anyscale version.",
)
@click.option(
    "--json",
    "show_json",
    is_flag=True,
    default=False,
    help="Return output as json, for use with --version.",
)
@click.pass_context
def cli(ctx: Any, version_flag: bool, show_json: bool) -> None:
    # Use anyscale vendored Ray for the CLI:
    sys.path.insert(0, anyscale.ANYSCALE_RAY_DIR)
    if (
        (ctx.invoked_subcommand not in {"help", "version"})
        and ("--help" not in sys.argv)
        and ("-h" not in sys.argv)
    ):
        init_sentry(ctx.invoked_subcommand)

    if version_flag:
        ctx.invoke(version_cli, show_json=show_json)


@click.command(name="version", help="Display version of the anyscale CLI.")
@click.option(
    "--json", "show_json", is_flag=True, default=False, help="Return output as json."
)
def version_cli(show_json: bool) -> None:
    if show_json:
        print(json.dumps({"version": anyscale.__version__}))
    else:
        log.info(anyscale.__version__)


@cli.command(
    name="help", help="Display help documentation for anyscale CLI.", hidden=True
)
@click.pass_context
def anyscale_help(ctx: Any) -> None:
    print(ctx.parent.get_help())


cli.add_command(session_cli)
cli.add_command(cloud_cli)
cli.add_command(config_cli)
cli.add_command(migrate_cli)
cli.add_command(project_cli)
cli.add_command(version_cli)
cli.add_command(list_cli)
cli.add_command(cluster_env_cli)
cli.add_command(job_cli)
cli.add_command(service_cli)
cli.add_command(cluster_cli)

cli.add_command(anyscale_init)
cli.add_command(anyscale_up)
cli.add_command(anyscale_stop)
cli.add_command(anyscale_autosync)
cli.add_command(anyscale_autopush)
cli.add_command(anyscale_fork)
cli.add_command(anyscale_clone)
cli.add_command(anyscale_ssh)
cli.add_command(anyscale_start)
cli.add_command(anyscale_exec)
cli.add_command(anyscale_run)
cli.add_command(anyscale_kill_job)
cli.add_command(anyscale_pull)
cli.add_command(anyscale_push)
cli.add_command(anyscale_help)
cli.add_command(cluster_compute_cli)

# Commands to interact with the Anyscale API
cli.add_command(anyscale_api)

ALIASES = {"h": anyscale_help}


def main() -> Any:
    return cli()


if __name__ == "__main__":
    main()
